""" Time based config store

This config store updates the configuration every x seconds
"""
import threading
import time

from boto.dynamodb2.exceptions import ItemNotFound

from dynamodb_config_store.config_stores import ConfigStore


class TimeBasedConfigStore(ConfigStore):
    """ Fetches Stores on a periodic interval from DynamoDB

    All internal variables and methods are private (with an leading _), as
    the configuration options from DynamoDB will be set as instance attributes.
    """

    _option_key = None      # Option key in DynamoDB
    _populated = False      # True when the first population has been done
    _store_key = None       # Store key in DynamoDB
    _store_name = None      # Name of the Store
    _table = None           # boto.dynamodb2.table.Table
    _update_interval = 300  # How often, in seconds, to fetch updates

    def __init__(
            self, table, store_name, store_key, option_key,
            update_interval=300):
        """ Constructor for the TimeBasedConfigStore

        :type table: boto.dynamodb2.table.Table
        :param table: Table instance
        :type store_name: str
        :param store_name: Name of the DynamoDB Config Store
        :type store_key: str
        :param store_key: Key name for the store in DynamoDB. Default _store
        :type option_key: str
        :param option_key: Key name for the option in DynamoDB. Default _option
        :type update_interval: int
        :param update_interval: How often, in seconds, to fetch updates
        :returns: None
        """
        self._option_key = option_key
        self._store_key = store_key
        self._store_name = store_name
        self._table = table
        self._update_interval = update_interval

        thread = threading.Thread(target=self._run, args=())
        thread.daemon = True
        thread.start()

        while not self._populated:
            time.sleep(0.5)

    def _run(self):
        """ Run periodic fetcher

        :returns: None
        """
        while True:
            # Get options from DynamoDB
            options = self._fetch_options()

            # Delete old attributes
            self._delete_instance_attributes()

            # Populate the attribute list with the new attributes
            self._attributes = [key for key in options.keys()]

            # Add new attributes
            self._set_instance_attributes(options)

            self._populated = True

            time.sleep(self._update_interval)

    def _fetch_options(self):
        """ Retrieve a dictionary with all options and values from DynamoDB

        :returns: dict -- Dict with {'option': {'key': 'value'}}
        """
        try:
            items = {}
            query = {'{}__eq'.format(self._store_key): self._store_name}

            for item in self._table.query_2(**query):
                option = item[self._option_key]

                # Remove metadata
                del item[self._store_key]
                del item[self._option_key]

                items[option] = {k: v for k, v in item.items()}

            return items

        except ItemNotFound:
            return {}



""" Config Store base class """


class ConfigStore(object):
    """ Base class for config stores """

    _attributes = []        # List of set instance attributes

    def _delete_instance_attributes(self):
        """ Delete all the instance attributes

        :returns: None
        """
        for attr in self._attributes:
            delattr(self, attr)

    def _set_instance_attributes(self, options):
        """ Set instance attributes

        :type options: dict
        :param options: Dict with {'option': {'key': 'value'}}
        :returns: None
        """
        [setattr(self, key, options[key]) for key in options.keys()]



""" DynamoDB Config Store

                    APACHE LICENSE 2.0
            Copyright 2014 Sebastian Dahlgren

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Example structure:

------------+----------------+----------------+----------------+----------------
_store*     | _option**      | host           | port           | secret-key
------------+----------------+----------------+----------------+----------------
prod        | db             | db-cluster.com | 27017          |
prod        | external-port  |                | 80             |
prod        | secret-key     |                |                | abcd1234
test        | db             | localhost      | 27017          |
test        | external-port  |                | 4000           |
prod        | secret-key     |                |                | test1234
------------+----------------+----------------+----------------+----------------

*) Hash key
**) Range key
"""
import os.path
import time
import sys
if sys.version_info.major > 2:
    from configparser import ConfigParser as SafeConfigParser
else:
    from ConfigParser import SafeConfigParser

from boto.dynamodb2.exceptions import (
    LimitExceededException,
    ProvisionedThroughputExceededException,
    ResourceInUseException,
    ResourceNotFoundException,
    ValidationException)
from boto.dynamodb2.fields import HashKey, RangeKey
from boto.dynamodb2.table import Table
from boto.exception import JSONResponseError

from dynamodb_config_store.config_stores.simple import SimpleConfigStore
from dynamodb_config_store.config_stores.time_based import TimeBasedConfigStore
from dynamodb_config_store.exceptions import (
    MisconfiguredSchemaException,
    TableNotCreatedException,
    TableNotReadyException)

# Publish the module __version__
config_file = SafeConfigParser()
config_file.read(
    os.path.abspath(os.path.join(os.path.dirname(__file__), 'settings.conf')))
__version__ = config_file.get('general', 'version')


class DynamoDBConfigStore(object):
    """ DynamoDB Config Store instance """

    config = None           # Instance of the a ConfigStore
    connection = None       # boto.dynamodb2.layer1.DynamoDBConnection instance
    option_key = None       # Key for the option (default: _option)
    read_units = None       # Number of read units to provision to new tables
    store_key = None        # Key for the store (default: _store)
    store_name = None       # Name of the Store
    config_store = None       # Store type to use
    config_store_args = None  # Store type arguments
    config_store_kwargs = None  # Store type key word args
    table = None            # boto.dynamodb2.table.Table instance
    table_name = None       # Name of the DynamoDB table
    write_units = None      # Number of write units to provision to new tables

    def __init__(
            self, connection, table_name, store_name,
            store_key='_store', option_key='_option',
            read_units=1, write_units=1,
            config_store='SimpleConfigStore',
            config_store_args=[], config_store_kwargs={}):
        """ Constructor for the config store

        :type connection: boto.dynamodb2.layer1.DynamoDBConnection
        :param connection: Boto connection object to use
        :type table_name: str
        :param table_name: Name of the DynamoDB table to use
        :type store_name: str
        :param store_name: Name of the DynamoDB Config Store
        :type store_key: str
        :param store_key: Key name for the store in DynamoDB. Default _store
        :type option_key: str
        :param option_key: Key name for the option in DynamoDB. Default _option
        :type config_store: str
        :param config_store: Store type to use
        :type config_store_args: list
        :param config_store_args: Store type arguments
        :type config_store_kwargs: dict
        :param config_store_kwargs: Store type key word arguments
        :returns: None
        """
        self.connection = connection
        self.option_key = option_key
        self.read_units = read_units
        self.store_key = store_key
        self.store_name = store_name
        self.table_name = table_name
        self.write_units = write_units
        self.config_store = config_store
        self.config_store_args = config_store_args
        self.config_store_kwargs = config_store_kwargs

        self._initialize_table()
        self._initialize_store()

    def _initialize_store(self):
        """ Initialize the store to use """
        if self.config_store == 'TimeBasedConfigStore':
            self.config = TimeBasedConfigStore(
                self.table,
                self.store_name,
                self.store_key,
                self.option_key,
                *self.config_store_args,
                **self.config_store_kwargs)
        elif self.config_store == 'SimpleConfigStore':
            self.config = SimpleConfigStore(
                self.table,
                self.store_name,
                self.store_key,
                self.option_key,
                *self.config_store_args,
                **self.config_store_kwargs)
        else:
            raise NotImplementedError

    def _initialize_table(self):
        """ Initialize the table

        :returns: None
        """
        try:
            table = self.connection.describe_table(self.table_name)
            status = table[u'Table'][u'TableStatus']
            schema = table[u'Table'][u'KeySchema']

            # Validate that the table is in ACTIVE state
            if status not in ['ACTIVE', 'UPDATING']:
                raise TableNotReadyException

            # Validate schema
            hash_found = False
            range_found = False
            for key in schema:
                if key[u'AttributeName'] == self.store_key:
                    if key[u'KeyType'] == u'HASH':
                        hash_found = True

                if key[u'AttributeName'] == self.option_key:
                    if key[u'KeyType'] == u'RANGE':
                        range_found = True

            if not hash_found or not range_found:
                raise MisconfiguredSchemaException

        except JSONResponseError as error:
            if error.error_code == 'ResourceNotFoundException':
                table_created = self._create_table(
                    read_units=self.read_units,
                    write_units=self.write_units)

                if not table_created:
                    raise TableNotCreatedException

        self.table = Table(self.table_name, connection=self.connection)

    def _create_table(self, read_units=1, write_units=1):
        """ Create a new table

        :type read_units: int
        :param read_units: Number of read capacity units to provision
        :type write_units: int
        :param write_units: Number of write capacity units to provision
        :returns: bool -- Returns True if the table was created
        """
        self.table = Table.create(
            self.table_name,
            schema=[
                HashKey(self.store_key),
                RangeKey(self.option_key)
            ],
            throughput={
                'read': read_units,
                'write': write_units
            },
            connection=self.connection)

        # Wait for the table to get ACTIVE
        return self._wait_for_table(target_state='ACTIVE')

    def _wait_for_table(self, target_state, sleep_time=5, retries=30):
        """ Wait for the table to get to a certain state

        :type target_state: str
        :param target_state: The target state to wait for
        :type sleep_time: int
        :param sleep_time: Number of seconds to wait between the checks
        :type retries: int
        :param retries: Number of retries before giving up
        :returns: bool -- True if the target state was reached, else False
        """
        while retries > 0:
            desc = self.connection.describe_table(self.table_name)
            if desc[u'Table'][u'TableStatus'] == target_state.upper():
                return True

            time.sleep(sleep_time)
            retries -= 1

        return False

    def reload(self):
        """ Reload the config store

        :returns: None
        """
        self._initialize_store()

    def set(self, option, data):
        """ Upsert a config item

        A write towards DynamoDB will be executed when this method is called.

        :type option: str
        :param option: Name of the configuration option
        :type data: dict
        :param data: Dictionary with all option data
        :returns: bool -- True if the data was stored successfully
        """
        data[self.store_key] = self.store_name
        data[self.option_key] = option

        try:
            return self.table.put_item(data, overwrite=True)
        except LimitExceededException:
            raise
        except ProvisionedThroughputExceededException:
            raise
        except ResourceInUseException:
            raise
        except ResourceNotFoundException:
            raise
        except ValidationException:
            raise
        except Exception:
            raise



""" The Simple Config Store implementation """
from boto.dynamodb2.exceptions import ItemNotFound

from dynamodb_config_store.config_stores import ConfigStore


class SimpleConfigStore(ConfigStore):
    """ SimpleConfigStore fetching configuration directly from DynamoDB

    This config store will always poll for the latest changes from DynamoDB.
    """

    _option_key = None      # Option key in DynamoDB
    _populated = False      # True when the first population has been done
    _store_key = None       # Store key in DynamoDB
    _store_name = None      # Name of the Store
    _table = None           # boto.dynamodb2.table.Table

    def __init__(self, table, store_name, store_key, option_key):
        """ Constructor for the SimpleConfigStore

        :type table: boto.dynamodb2.table.Table
        :param table: Table instance
        :type store_name: str
        :param store_name: Name of the DynamoDB Config Store
        :type store_key: str
        :param store_key: Key name for the store in DynamoDB. Default _store
        :type option_key: str
        :param option_key: Key name for the option in DynamoDB. Default _option
        :returns: None
        """
        self._option_key = option_key
        self._store_key = store_key
        self._store_name = store_name
        self._table = table

    def get(self, option=None, keys=None):
        """ Get a config item

        A query towards DynamoDB will always be executed when this
        method is called.

        An boto.dynamodb2.exceptions.ItemNotFound will be thrown if the config
        option does not exist.

        :type option: str
        :param option: Name of the configuration option, all options if None
        :type keys: list
        :param keys: List of keys to return (used to get subsets of keys)
        :returns: dict -- Dictionary with all data; {'key': 'value'}
        """
        if option:
            return self.get_option(option, keys=keys)

        else:
            try:
                items = {}
                query = {'{}__eq'.format(self._store_key): self._store_name}

                for item in self._table.query_2(**query):
                    option = item[self._option_key]

                    # Remove metadata
                    del item[self._store_key]
                    del item[self._option_key]

                    items[option] = {k: v for k, v in item.items()}

                return items

            except ItemNotFound:
                raise

    def get_option(self, option, keys=None):
        """ Get a specific option from the store.

        A query towards DynamoDB will always be executed when this
        method is called.

        get_option('a') == get(option='a')
        get_option('a', keys=['b', 'c']) == get(option='a', keys=['b', 'c'])

        :type option: str
        :param option: Name of the configuration option
        :type keys: list
        :param keys: List of keys to return (used to get subsets of keys)
        :returns: dict -- Dictionary with all data; {'key': 'value'}
        """
        try:
            kwargs = {
                self._store_key: self._store_name,
                self._option_key: option
            }

            item = self._table.get_item(**kwargs)

            del item[self._store_key]
            del item[self._option_key]

            if keys:
                return {
                    key: value
                    for key, value in item.items()
                    if key in keys
                }
            else:
                return {key: value for key, value in item.items()}

        except ItemNotFound:
            raise



""" Setup script for PyPI """
import os
from setuptools import setup
import sys
if sys.version_info.major > 2:
    from configparser import ConfigParser as SafeConfigParser
else:
    from ConfigParser import SafeConfigParser

settings = SafeConfigParser()
settings.read(os.path.realpath('dynamodb_config_store/settings.conf'))

setup(
    name='dynamodb-config-store',
    version=settings.get('general', 'version'),
    license='Apache License, Version 2.0',
    description='Store configuration details in DynamoDB',
    author='Sebastian Dahlgren',
    author_email='sebastian.dahlgren@gmail.com',
    url='https://github.com/sebdah/dynamodb-config-store/',
    keywords="dynamodb aws config configuration amazon web services",
    platforms=['Any'],
    packages=['dynamodb_config_store'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'boto>=2.29.1'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ]
)



""" Exception definitions """


class TableNotCreatedException(Exception):
    """ Exception thrown if the table could not be created successfull """
    pass


class TableNotReadyException(Exception):
    """ Exception thrown if the table is not in ACTIVE or UPDATING state """
    pass


class MisconfiguredSchemaException(Exception):
    """ Exception thrown if the table does not match the configuration """
    pass
